@@ -2,6 +2,5 @@
 from manifest.manifest import Manifest
 from manifest.request import Request
 from manifest.response import Response
-from manifest.session import Session
 
-__all__ = ["Manifest", "Response", "Session"]
+__all__ = ["Manifest", "Response", "Request"]
