@@ -1,14 +1,13 @@
 """Cache for queries and responses."""
 from abc import ABC, abstractmethod
-from typing import Any, Dict, Union
+from typing import Any, Dict, Type, Union
 
-from manifest.caches.serializers import ArraySerializer, Serializer
-from manifest.response import RESPONSE_CONSTRUCTORS, Response
+from manifest.caches.serializers import ArraySerializer, NumpyByteSerializer, Serializer
+from manifest.request import DiffusionRequest, EmbeddingRequest, LMRequest, Request
+from manifest.response import Response
 
-CACHE_CONSTRUCTOR = {
-    "diffuser": ArraySerializer,
-    "tomadiffuser": ArraySerializer,
-}
+# Non-text return type caches
+ARRAY_CACHE_TYPES = {EmbeddingRequest, DiffusionRequest}
 
 
 class Cache(ABC):
@@ -17,39 +16,58 @@ class Cache(ABC):
     def __init__(
         self,
         connection_str: str,
-        client_name: str = "None",
+        request_type: Type[Request] = LMRequest,
         cache_args: Dict[str, Any] = {},
     ):
         """
-        Initialize client.
+        Initialize cache.
 
         Args:
             connection_str: connection string.
-            client_name: name of client.
+            request_type: request type.
             cache_args: arguments for cache.
 
-        cache_args are passed to client as default parameters.
+        cache_args are any arguments needed to initialize the cache.
 
-        For clients like OpenAI that do not require a connection,
-        the connection_str can be None.
+        Further, cache_args can contain `array_serializer` as a string
+        for embedding or image return types (e.g. diffusers) with values
+        as `local_file` or `byte_string`. `local_file` will save the
+        array in a local file and cache a pointer to the file.
+        `byte_string` will convert the array to a byte string and cache
+        the entire byte string. `byte_string` is default.
 
         Args:
-            connection_str: connection string for client.
+            connection_str: connection string for cache.
             cache_args: cache arguments.
         """
-        self.client_name = client_name
+        self.request_type = request_type
         self.connect(connection_str, cache_args)
-        self.serializer = CACHE_CONSTRUCTOR.get(client_name, Serializer)()
+        if self.request_type in ARRAY_CACHE_TYPES:
+            array_serializer = cache_args.pop("array_serializer", "byte_string")
+            if array_serializer not in ["local_file", "byte_string"]:
+                raise ValueError(
+                    "array_serializer must be local_file or byte_string,"
+                    f" not {array_serializer}"
+                )
+            self.serializer = (
+                ArraySerializer()
+                if array_serializer == "local_file"
+                else NumpyByteSerializer()
+            )
+        else:
+            # If user has array_serializer type, it will throw an error as
+            # it is not recognized for non-array return types.
+            self.serializer = Serializer()
 
     @abstractmethod
     def close(self) -> None:
-        """Close the client."""
+        """Close the cache."""
         raise NotImplementedError()
 
     @abstractmethod
     def connect(self, connection_str: str, cache_args: Dict[str, Any]) -> None:
         """
-        Connect to client.
+        Connect to cache.
 
         Args:
             connection_str: connection string.
@@ -101,14 +119,9 @@ class Cache(ABC):
         key = self.serializer.request_to_key(request)
         cached_response = self.get_key(key)
         if cached_response:
-            cached = True
             response = self.serializer.key_to_response(cached_response)
-            return Response(
-                response,
-                cached,
-                request,
-                **RESPONSE_CONSTRUCTORS.get(self.client_name, {})
-            )
+            response["cached"] = True
+            return Response.from_dict(response, request_dict=request)
         return None
 
     def set(self, request: Dict, response: Dict) -> None:
