@@ -44,9 +44,9 @@ class TOMADiffuserClient(TOMAClient):
         Returns:
             model params.
         """
-        return {"model_name": "tomadiffuser", "engine": getattr(self, "engine")}
+        return {"model_name": self.NAME, "engine": getattr(self, "engine")}
 
-    def format_response(self, response: Dict, request: Dict) -> Dict[str, Any]:
+    def validate_response(self, response: Dict, request: Dict) -> Dict[str, Any]:
         """
         Format response to dict.
 
