@@ -3,12 +3,14 @@ from typing import Any, Dict, List, Optional, Tuple, Union
 
 from pydantic import BaseModel
 
+# Used when unioning requests after async connection pool
+ENGINE_SEP = "::"
 NOT_CACHE_KEYS = {"client_timeout", "batch_size"}
+# The below should match those in Request.
 DEFAULT_REQUEST_KEYS = {
     "client_timeout": ("client_timeout", 60),  # seconds
-    "batch_size": ("batch_size", 1),
+    "batch_size": ("batch_size", 8),
     "run_id": ("run_id", None),
-    "request_type": ("request_type", None),
 }
 
 
@@ -25,7 +27,7 @@ class Request(BaseModel):
     n: int = 1
 
     # Timeout
-    client_timeout: int = 120
+    client_timeout: int = 60
 
     # Run id used to repeat run with same parameters
     run_id: Optional[str] = None
@@ -33,15 +35,15 @@ class Request(BaseModel):
     # Batch size for async batch run
     batch_size: int = 8
 
-    # Request type None is for completion. Used to scoring prompt
-    request_type: str = None
-
     def to_dict(
         self, allowable_keys: Dict[str, Tuple[str, Any]] = None, add_prompt: bool = True
     ) -> Dict[str, Any]:
         """
         Convert request to a dictionary.
 
+        Handles parameter renaming but does not fill in default values.
+        It will drop any None values.
+
         Add prompt ensures the prompt is always in the output dictionary.
         """
         if allowable_keys:
@@ -74,6 +76,9 @@ class LMRequest(Request):
     # Top k sampling taking top_k highest probability tokens
     top_k: int = 50
 
+    # Logprobs return value
+    logprobs: Optional[int] = None
+
     # Stop sequences
     stop_sequences: Optional[List[str]] = None
 
@@ -96,12 +101,27 @@ class LMRequest(Request):
     frequency_penalty: float = 0
 
 
+class LMChatRequest(LMRequest):
+    """Language Model Chat Request object."""
+
+    prompt: List[Dict[str, str]] = {}  # type: ignore
+
+
+class LMScoreRequest(LMRequest):
+    """Language Model Score Request object."""
+
+    pass
+
+
+class EmbeddingRequest(Request):
+    """Embedding Request object."""
+
+    pass
+
+
 class DiffusionRequest(Request):
     """Diffusion Model Request object."""
 
-    # Request type
-    request_type: str = "diffusion"
-
     # Number of steps
     num_inference_steps: int = 50
 
